{{
=====================================================================
SdramCache.spin - defines a device driver for ISSI 32Mx8 SDRAM
=====================================================================

    Change History
    --------------
    Oct 25, 2010: Initial release for use with SdramTest.spin
    Nov  5, 2010: Added dirty write-back handling for better read performance.
    Nov  6, 2010: Reverted an aggressive optimization - performance is same.
    Mar 31, 2011: [Tom Dinger] Revised (significantly) the logic and
                  scheduling around refreshes, and the calculation of
                  timing; moved the init code into the tag memory, and
                  increased the maximum number of tags to 128.
}}

{{
    Copyright (c) 2010 by John Steven Denson (jazzed)
    Modifications (c) Copyright 2011 Thomas Dinger (Ding-Batty)

    TERMS OF USE: MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
    THE SOFTWARE.
}}

{{
=====================================================================
Any memory interface can be used given enough pins; a parallel
bus interface is challenging since Propeller only has 32 pins.
An 8 bit data bus is a reasonable trade for pin use/performance.
With a cache, the a bus width > 8 bits is best if possible.

This SDRAM solution uses 20 propeller pins.
for A0-12, BA0-1, CS*, CLK, RAS*, CAS*, D0-7 (CKE is pulled high).

The CS* and DQM pins are permanently pulled low for a 32Mx8 memory.

D0-7       on A0-7 via ALE on P15
A8-14      on A8-14
SDRAM ALE  on P21
SDRAM CLK  on P22
SDRAM RAS  on P25
SDRAM CAS  on P26
SDRAM WE*  on P27

With this design, there are 8 Propeller pins (P15-20, P23-24) free for other IO.
Current pin usage: P0-14 & P21-22 & P25-27 SDRAM, P28-29 I2C, P30-31 Serial Port.
=====================================================================
}}


{{
=====================================================================
SDRAM CACHE device driver code
=====================================================================

This driver must be built with a compiler that supports #ifdef
In BST/BSTC "non-parallax extentions" must be enabled for builds.

The options available with #ifdef are:

LINELEN_32      ' use  32 byte cache line size - slowest burst rate.
LINELEN_64      ' use  64 byte cache line size
LINELEN_128     ' use 128 byte cache line size
LINELEN_256     ' use 256 byte cache line size
LINELEN_512     ' use 512 byte cache line size - fastest possible burst rate

LINELEN is inversely proportional to the number of cache lines in use.
Some applications such as a C program will benefit from smaller LINELEN.
Video or other buffer application need larger a LINELEN.

Total cache size in HUB memory, in bytes, is controlled by selecting one
of the following #defines:

CACHE_16K
CACHE_8K
CACHE_4K
CACHE_2K
CACHE_1K

These defines allow smaller caches for programs needing more HUB
resources. If cache size is not defined, the cache size is 2KB.

The number of cache lines is determined by the combination of LINELEN and
CACHE values. The following table determines how many cache lines are
available for each size, to a maximum of 128 cache lines:

           LINELEN
              32       64      128      256      512
            -----    -----    -----    -----    ----- 
CACHE_16K      x        x      128       64       32
CACHE_8K       x      128       64       32       16
CACHE_4K      128      64       32       16        8
CACHE_2K      64       32       16        8        4
CACHE_1K      32       16        8        4        2

The #defines are set in the driver, but can be set in another file.
}}

{{
=====================================================================
These constants define the pin functions:
}}
con
    DATABUS_MASK  = $FF
    ADDRBUS_MASK  = $7FFF

    ALE_PIN       = 15          ' TODO: Move to near the CLK pin (P21)

    RAMCLK_PIN    = 24          ' TODO: Move out of P23..P31 so we can use
                                ' MOVI to set RAS, CAS and WE (set the command)
                                ' So move it to P22
    RASN_PIN      = 25
    CASN_PIN      = 26
    WEN_PIN       = 27

    RASN_MASK   = (|< RASN_PIN)
    CASN_MASK   = (|< CASN_PIN)
    WEN_MASK    = (|<  WEN_PIN)
    DCMD_MASK   = RASN_MASK | CASN_MASK | WEN_MASK
    RAMCLK_MASK = (|< RAMCLK_PIN)
    ALE_MASK    = (|< ALE_PIN)


    FREE_MASK   = !(DCMD_MASK | RAMCLK_MASK | ALE_MASK | ADDRBUS_MASK)
                                ' unused/free pin mask

    ' Constants for processing address parameters
    ADDR_MASK        = (|< 25)-1   ' 32MB
    ROW_MASK         = (|< 13)-1   ' A0-12
    COL_MASK         = (|< 10)-1   ' A0-9
    CASN_PREPIN_MASK = (|< 10)     ' A10 is precharge pin

    MODE_REG         = $37           ' CAS latency 3, full page burst
                                     ' TODO: Investigate CAS latency 2...

{{
=====================================================================
These constants define the driver commands.
}}
con
    CMD_MASK    = $3
    READ_CMD    = $3
    WRITE_CMD   = $2
    DIRTY_BIT   = $1

con
{{
=====================================================================
Cache is organized as a table.
The tagline table contains pointers to cache entries.
Each tagline has the following definition:

    TAGLINE:    xxxx_xxxi
    such that:  x=match address i=ignored bits

The actual number of ignored bits depends on the cache line length.
}}

#define USE_DIRTY       ' allows cache write to flush only when necessary

#define LINELEN_32
'#define LINELEN_64      ' fastest reload over broad range of addresses
'#define LINELEN_128
'#define LINELEN_256
'#define LINELEN_512     ' best for linear memory test and read-only bursts

#ifndef LINELEN_32
#ifndef LINELEN_64
#ifndef LINELEN_128
#ifndef LINELEN_256
#ifndef LINELEN_512
#define LINELEN_32       ' The default, if nothing else defined.
#endif
#endif
#endif
#endif
#endif


'#define CACHE_16K
'#define CACHE_8K
'#define CACHE_4K
#define CACHE_2K
'#define CACHE_1K

#ifndef CACHE_1K
#ifndef CACHE_2K
#ifndef CACHE_4K
#ifndef CACHE_8K
#ifndef CACHE_16K
#define CACHE_2K         ' The default, if nothing else defined.
#endif
#endif
#endif
#endif
#endif

' Now, for each pairing, define the actual values that will be used.
#ifdef     CACHE_1K
    CACHESIZE_BYTES =  1 * 1024
#ifdef       LINELEN_32
    LINESHIFT       =  5
#elseifdef   LINELEN_64
    LINESHIFT       =  6
#elseifdef   LINELEN_128
    LINESHIFT       =  7
#elseifdef   LINELEN_256
    LINESHIFT       =  8
#elseifdef   LINELEN_512
    LINESHIFT       =  9
#endif
#elseifdef CACHE_2K
    CACHESIZE_BYTES =  2 * 1024
#ifdef       LINELEN_32
    LINESHIFT       =  5
#elseifdef   LINELEN_64
    LINESHIFT       =  6
#elseifdef   LINELEN_128
    LINESHIFT       =  7
#elseifdef   LINELEN_256
    LINESHIFT       =  8
#elseifdef   LINELEN_512
    LINESHIFT       =  9
#endif
#elseifdef CACHE_4K
    CACHESIZE_BYTES =  4 * 1024
#ifdef       LINELEN_32
    LINESHIFT       =  5
#elseifdef   LINELEN_64
    LINESHIFT       =  6
#elseifdef   LINELEN_128
    LINESHIFT       =  7
#elseifdef   LINELEN_256
    LINESHIFT       =  8
#elseifdef   LINELEN_512
    LINESHIFT       =  9
#endif
#elseifdef CACHE_8K
    CACHESIZE_BYTES =  8 * 1024
#ifdef       LINELEN_32
#error ERROR: Unsupported combination: 32 byte cache lines for 8K cache
#elseifdef   LINELEN_64
    LINESHIFT       =  6
#elseifdef   LINELEN_128
    LINESHIFT       =  7
#elseifdef   LINELEN_256
    LINESHIFT       =  8
#elseifdef   LINELEN_512
    LINESHIFT       =  9
#endif
#elseifdef CACHE_16K
    CACHESIZE_BYTES = 16 * 1024
#ifdef       LINELEN_32
#error ERROR: Unsupported combination: 32 byte cache lines for 16K cache
#elseifdef   LINELEN_64
#error ERROR: Unsupported combination: 64 byte cache lines for 16K cache
#elseifdef   LINELEN_128
    LINESHIFT       =  7
#elseifdef   LINELEN_256
    LINESHIFT       =  8
#elseifdef   LINELEN_512
    LINESHIFT       =  9
#endif
#endif

    LINESHFT        = LINESHIFT     ' spelling variation

    LINELEN_BYTES   = |< LINESHIFT  ' Bytes in one cache line
    CACHESIZE_LONGS = CACHESIZE_BYTES >> 2
 
    LINELEN         = LINELEN_BYTES

    TAGCOUNT        = CACHESIZE_BYTES/LINELEN_BYTES ' cache with linelen bytes
                                    ' from 2 to 128 cache lines.
    MAX_TAGCOUNT    = 128           ' for tag vector storage allocation

    TAGMASK         = $003F_FFFF    ' $3FFFFF*512 = 2GB $3FFFFF*32 = 128MB
    DIRTYSHFT       = 31            ' shift by N to get dirty bit   test tag,DIRTYMASK wc
    DIRTYMASK       = |< DIRTYSHFT  ' dirty bit mask                muxc tag,DIRTYSHFT

    ' Define some more constants, for use in Refresh calculations
    ' NOTE: These really are interdependent: the number of refreshes done
    ' in a single burst determines the maximum ("optimal") number of bytes
    ' that can be transferred (read or written) between the refresh bursts,
    ' and the refresh loop count limit is determined by:
    ' - the clock rate
    ' - the number of rows in the SDRAM to be refreshed (8192 here)
    ' - the total time to refresh all rows (64 ms here)
    ' - the number of refreshes done in each burst
    ' - the cost of performing the refresh burst
    '
    ' Now, the minimum refresh burst count is determined by
    ' the number of bytes in a block transfer, which determines
    ' how many clocks between refresh bursts:
    '   xferclks = (clksperbyte * bytes) + xferovrhd + refreshovrhd + (clksperrefr * refrs)
    '
    ' Refresh calculations:
    ' The approach taken here is that we have to do 8192 refresh commands
    ' in 64 ms. Each refresh command will take a certain amount of time, and the
    ' remaining time in the 64 ms is the available time for I/O transfers
    ' between the cog/hub and the SDRAM.
    '
    ' There is overhead for setting up each transfer: calculating and latching
    ' the SDRAM address, calculating and checking the cache, setting up
    ' for a burst of refresh commands. So, the larger the block of bytes in
    ' a single transfer, the higher the overall I/O throughput.
    '
    ' This all leads to issuing SDRAM refresh commands in bursts, so that
    ' the time between the bursts can  be made longer, to permit larger
    ' I/O blocks to be transferred.
    '
    ' There is a trade-off here in performance: When the I/O block is made
    ' larger, the time to read and write it is larger, which increases the
    ' latency until data is available when the desired data is not present
    ' in the cache. But larger I/O blocks means greater throughput, which
    ' can be important in some applications, such as those that are more
    ' like "streaming" the data from the SDRAM at high speeds. Furthermore,
    ' when the I/O block is larger, the number of refreshes in one refresh
    ' burst between the I/O transfers is also larger, making the I/O
    ' latency larger still.

    ' Our operating conditions: CHANGE MY_CLKFREQ AS NEEDED!
    ' This should match the value for the overall program using this
    ' SDRAM cache object.
    '
    ' TODO: Do the calculations for REFRESH_BURST_COUNT and
    ' CMDPOLL_LOOPS_PER_REFRESH_BURST at runtime, in the initialization
    ' code -- those are the only two settings that depend on the system
    ' clock frequency.

    MY_CLKFREQ                   = 80_000_000
    CLOCKS_IN_64MS               = (MY_CLKFREQ * 8) / 125 ' == (MY_CLKFREQ / 1000) * 64

    ' Some clock counts for the implementation:
    ' CLKS_PER_BYTE -- the number of system clocks per byte transferred.
    ' CLKS_OVHD_CMD -- the common overhead for command dispatch, either read or write.
    ' CLKS_OVHD_READ -- the number of system clocks in overhead for one
    '                   block read operation.
    ' CLKS_OVHD_WRITE -- the number of system clocks in overhead for one
    '                    block write operation.
    ' CLKS_PER_REFR -- the number of system clocks for each REFRESH command
    '                  done in the sdram_refresh routine.
    ' CLKS_OVHD_REFR -- the number of system clocks in overhead for running
    '                   the sdram_refresh routine

    ' Make these even more accurate once the "final" I/O routines
    ' are done.

    CLKS_PER_BYTE   =  12
    CLKS_PER_REFR   =   8
    CLKS_OVHD_CMD   = 180
    CLKS_OVHD_READ  = 164 ' not counting cmd dispatch
    CLKS_OVHD_WRITE = 188 ' not counting cmd dispatch
    CLKS_OVHD_REFR  =  52

    ' This computes a max(read,write) function:
    CLKS_OVHD_RW = (((CLKS_OVHD_READ <  CLKS_OVHD_WRITE) & CLKS_OVHD_WRITE) {
}                 | ((CLKS_OVHD_READ => CLKS_OVHD_WRITE) & CLKS_OVHD_READ ))

    XFERCLKS_WO_REFR = (CLKS_PER_BYTE * LINELEN) + CLKS_OVHD_RW + CLKS_OVHD_CMD
    IOCLKS           = XFERCLKS_WO_REFR + CLKS_OVHD_REFR

    ' We know we have to do 8192 refreshes within 64 ms, and we know that they
    ' will take (8192 * CLKS_PER_REFR) for the refreshes, so the remaining clks
    ' will be for I/O transfer and overhead for I/O.

    IOCLKS_IN_64MS = CLOCKS_IN_64MS - (8192 * CLKS_PER_REFR)

    ' Now, for each transfer, we need IOCLKS clocks (per-byte clks plus block
    ' overhead plus refresh burst setup overhead), so calculate how many of
    ' these fit entirely into the remaining time. That determines the maximum
    ' number of block transfers we can do in 64MS.

    IOBLKS_IN_64MS = IOCLKS_IN_64MS / IOCLKS

    ' We have to do 8192 refreshes in this many I/O blocks, which
    ' determines the number of refreshes per block
    ' This formulation rounds up if it does not divide evenly, using integer
    ' division.

  ' REFR_BURST_CNT_CALC = (8192 + IOBLKS_IN_64MS - 1) / IOBLKS_IN_64MS
    REFR_BURST_CNT_CALC = (8191 / IOBLKS_IN_64MS) + 1

    ' We need to know how many 16-clock loops we can do before we have to
    ' perform a refresh burst. This is calculated as:

    CLOCKS_FOR_1_REFR            = CLOCKS_IN_64MS / IOBLKS_IN_64MS

    ' Now, from this, we have to subtract the overhead of the refresh routine,
    ' and then divide by 16 for the command polling loop time.

    CALCCLOCKS_IN_REFR_BURST_FN   = CLKS_OVHD_REFR + (CLKS_PER_REFR * REFR_BURST_CNT_CALC)
    CALCPOLL_LOOPS_PER_REFR_BURST = (CLOCKS_FOR_1_REFR - CALCCLOCKS_IN_REFR_BURST_FN) / 16

    ' Set the actual values we will use.
  ' CMDPOLL_LOOPS_PER_REFR_BURST = 625
  ' REFRESH_BURST_COUNT = 4
    CMDPOLL_LOOPS_PER_REFR_BURST = CALCPOLL_LOOPS_PER_REFR_BURST
    REFRESH_BURST_COUNT = REFR_BURST_CNT_CALC

    ' Viola! The "right" constant. (Whew!)

{{
=====================================================================
Driver mailbox interface - these longs must stay in this order.
}}
    ' These variables make up the parameter block passed to
    ' the PASM code at initialization. They contain:
    ' - the address of the command mailbox;
    ' - the address of the cache line within the cache memory block
    '   in HUB memory that contains the data of interest, calculated
    '   from the SDRAM address by the PASM code, and passed back to
    '   the caller issuing the command;
    ' - The base address of the block of HUB memory used for the cache.
    '
    ' This code used to supply a pointer to a block of HUB memory to
    ' store the 'tag' info (the per-cache-line info about the exact
    ' part of SDRAM currently in the cache), but that table is now
    ' in the COG itself.
    '
    ' Keep these in this order.

var
    long    mbcmd        ' Address of the HUB mailbox entry passing the
                         ' read/write command (and SDRAM address) to
                         ' the PASM code

    long    mbdata       ' Address of the HUB mailbox entry that gets the
                         ' HUB address of the cache line containing the
                         ' data of interest for the request.

    long    mbcache      ' base HUB address of the cache.

    long    cog          ' contains the COGID+1 of the running PASM cog.

' For controlling debugging.
'#define BMAD

{{
=====================================================================
start(mbptr)
mbptr = start address of command interface
Example:
var long command, data
pub main
  start(@command)
  repeat
}}
pub start(mbptr)
    stop                        ' Make sure we are not running already
    cache  := $AAAAAAAA         ' sets the first long in first line of cache.
    mbcmd  := mbptr
    mbdata := mbptr+4
'   mbtags := @tagram           ' tags are in-cog now
    mbcache:= @cache
    long[mbcmd]~~               ' set now so we know when startup is done

#ifdef BMAD
    bu.debug(@pasm,@mbcmd)      ' Add the debugging hooks, and start
                                ' the PASM cog with PASM debugging.

obj bu : "BMAUtility"           ' Include the PASM debugger.

#else
    cog := cognew(@pasm,@mbcmd)+1 ' just start the PASM cog
    repeat while long[mbcmd]    ' wait for startup to finish
    return cog                  ' start failed if return is 0
#endif


{{
=====================================================================
stop
}}
pub stop
    if cog                      ' if cog is zero, don't stop
        cogstop(cog~ - 1)       ' stop cog and set cog var to zero

{{
=====================================================================
dataptr - returns address of cache data buffer for the current operation
}}
pub dataptr
    return long[mbdata]

{{
=====================================================================
mboxptr  - returns address of cache command mailbox
}}
pub mboxptr 
    return long[mbcmd]

{{
=====================================================================
read a long
madr - long address to read
returns value read
NOTE: if the long address in SDRAM is not long-aligned, this may not return
the expected value.
}}
pub readLong(madr)
    long[mbcmd] := (madr & $FFFF_FFFC) | READ_CMD    ' issue read
    repeat while long[mbcmd]                         ' wait for I/O to complete
    return long[ long[mbdata] + (madr & (LINELEN-1)) ] ' get data from cache line

{{
=====================================================================
write a long
madr - long address to write
val  - value to write
NOTE: if the long address in SDRAM is not long-aligned, this may modify unexpected
memory locations.
}}
pub writeLong(madr,val)
    long[mbcmd] := (madr & $FFFF_FFFC) | WRITE_CMD   ' issue write
    repeat while long[mbcmd]                         ' wait for I/O to complete
    ' sending write command says this data will be written on next cache flush

    long[ long[mbdata] + (madr & (LINELEN-1)) ] := val

{{
=====================================================================
read a word
madr - word address to read
returns value read
NOTE: if the word address in SDRAM is not word-aligned, this may not return
the expected value.
}}
pub readWord(madr)
    long[mbcmd] := (madr & $FFFF_FFFC) | READ_CMD
    repeat while long[mbcmd]
    return word[ long[mbdata] + (madr & (LINELEN-1)) ]

{{
=====================================================================
write a word
madr - word address to write
val  - value to write
NOTE: if the word address in SDRAM is not word-aligned, this may modify unexpected
memory locations.
}}
pub writeWord(madr,val)
    long[mbcmd] := (madr & $FFFF_FFFC) | WRITE_CMD
    repeat while long[mbcmd]
    ' sending write command says this data will be written on next cache flush
    word[ long[mbdata] + (madr & (LINELEN-1)) ] := val

{{
=====================================================================
read a byte
madr - byte address to read
returns value read
}}
pub readByte(madr)
    long[mbcmd] := (madr & $FFFF_FFFC) | READ_CMD
    repeat while long[mbcmd]
    return byte[ long[mbdata] + (madr & (LINELEN-1)) ]

{{
=====================================================================
write a byte
madr - byte address to write
val  - value to write
}}
pub writeByte(madr,val)
    long[mbcmd] := (madr & $FFFF_FFFC) | WRITE_CMD
    repeat while long[mbcmd]
    ' sending write command says this data will be written on next cache flush
    byte[ long[mbdata] + (madr & (LINELEN-1)) ] := val

{{
=====================================================================
read buffer
NOTE: if the buffer address in SDRAM is not cache-line-length aligned,
this may not return the expected values.
}}
pub readBuffer(madr,datp) | n
    long[mbcmd] := (madr & $FFFF_FFFC) | READ_CMD
    repeat while long[mbcmd]
    longmove( datp, long[mbdata], LINELEN>>2 )

{{
=====================================================================
write buffer
NOTE: if the long address in SDRAM is not cache line length aligned,
this may modify unexpected memory locations.
}}
pub writeBuffer(madr,datp) | n
    long[mbcmd] := (madr & $FFFF_FFFC) | WRITE_CMD
    repeat while long[mbcmd]
    ' sending write command says this buffer will be written on next flush
    longmove(long[mbdata],datp,LINELEN>>2)

{{
=====================================================================
read cache tag line
madr - tag line address entry to read
returns value read
}}
pub readTags(madr)
    ' tagram is in cog now, so we're faking the address
    return (madr/constant(LINELEN>>2))<<LINESHIFT
'   return tagram[madr/constant(LINELEN>>2)]

{{
=====================================================================
read cache buffer
madr - cache address to read
returns value read
}}
pub readCache(madr)
    return cache[madr]

{{
=====================================================================
zero cache buffer
clears contents of the cache
}}
pub zeroCache
    ' bytefill(@cache,0,CACHESIZE_BYTES)
    ' if we know the cache is long-aligned, we could use:
    longfill(@cache,0, CACHESIZE_LONGS)

{{
=====================================================================
cache
}}
dat ' cache
'tagram          long 0 [TAGCOUNT]

' If the cache is at leat as big as the PASM code, we can reclaim
' all of the PASM code for use as cache memory in the HUB. But if we only
' want a 1K cache, then it fits _entirely_ within the PASM code space.
#ifdef     CACHE_1K
cache           long 0 ' the rest is the first part of the PASM code
#else
cache           long 0 [CACHESIZE_LONGS-$1e0]     ' reclaim PASM code for cache
#endif

{{
=====================================================================
PASM Cache Driver
=====================================================================
}}
dat                 org 0
pasm
'                    long    0 [8]           ' debugger stub space - or beginning of tag RAM

'--------------------------------------------------------------------
' startup and some variables
'
tagramincog                                 ' start of tag vector in COG
                    'jmp #sdram_setup        ' 0

                    ' The init code has been reworked, and is put almost
                    ' entirely within the tag vector. The init code
                    ' is less than 128 instructions, so it fits here.
                    ' After init, the tag vector is cleared, initializing
                    ' the tag vector and clobbering the one-time init.

sdram_setup
                    ' get mailbox pointers from the initialization param block:
                    ' - cmd mailbox address
                    ' - data mailbox address
                    ' - cache memory address in HUB
 
                    mov     tmp,par
                    rdlong  cmdptr,tmp
                    add     tmp,#4
                    rdlong  datptr,tmp
                    add     tmp,#4
                    rdlong  cacheptr,tmp

sdram_enable

' enable outputs
                    ' setup NOP command in output pins:
                    '    NOP == 1,1,1 RAS,CAS,WEN
                    or      outa,_DCMD_CLK_MASK   ' set NOP for startup, RAMCLK disabled

                    ' clear the address bits
                    ' outa &= ~$7FFF
                    andn    outa,_ADDRBUS_MASK    ' clear address bits

                    ' Everything set to output:
                    ' Address bus, data bus, ALE, RAS, CAS, WEN
                    mov     dira, OUTPUT_MASK

                    ' Force data bus to output as well
                    ' (Already taken care of above)
                    'or      dira,#DATABUS_MASK    ' dbus output for setup

' setup sdram clock
                    ' force clock off (CLK high)
                    ' (Already taken care of above)
                    'or      outa,_RAMCLK_MASK      ' clock starts "disabled"

                    ' Turn Clock pin to output forever
                    or      dira,_RAMCLK_MASK      ' turn on clock output forever

                    ' Set up the counter A
                    ' CTRMODE = %00100 << 26 -- NCO single-ended
                    ' PLLDIV  = %000   << 23 -- ignored in NCO mode
                    ' APIN    = RAMCLK_PIN
                    ' Counter running 16-clocks period (8 low, 8 high),
                    ' starting when turned on
                    ' FRQA := $20_00_00_00    80MHz / 8 == 10 MHz

                    mov     frqa,_FRQA_FAST     ' set access frequency
                    mov     ctra,_CTRA          ' start the counter

                    ' Because we set the OUTA[CLK] to 1, there will
                    ' be no clock pulses seen (yet)

' write SDRAM mode bits.
' DQM is pulled high on the board; it should be low - rework.
'

' send mode command sequence

send_initmode
                    ' necessary to stabalize SDRAM before first precharge
                    ' clear the OUTA[CLK] pin, to enable the clock pulses
                    ' This will start sending NOPs
                    andn    outa,_RAMCLK_MASK      ' turn on clock

                    ' Delay 100 us, which is clock rate / 10_000
                    ' Here we use a fixed count of $80_00, which
                    ' is 32768 clocks
                    ' for an 80MHz clock that works out to
                    '     32768 / 80MHz == 4096/10MHZ
                    ' about 410 us
                    ' [TD] But since we are already using the nominal Clock
                    ' rate above (MY_CLKFREQ), we can use that to calculate
                    ' a better count for the delay
                    mov     ndx, initdelayclks ' about 100us
                    add     ndx,cnt
                    waitcnt ndx,ndx
                    nop                        ' to make it exactly 16 clock + wait

                    ' disable the clock -- we don't care here about
                    ' subdividing a clock pulse, because this is
                    ' a multiple of 16 clocks since we started it.

                    or      outa,_RAMCLK_MASK      ' turn off clock

                    ' We can just call the sdram_refresh routine, with
                    ' a burst count of 12 instead of REFRESH_BURST_COUNT
                    mov     irefr_limit, #((12 - 1) << 1)
                    call    #sdram_refresh
                    mov     irefr_limit, #((REFRESH_BURST_COUNT-1) << 1)

                    ' TODO: Clean up this code...
                    ' Move all the logic for this here, into the tag
                    ' vector memory, to free it up once init is done.
                    call    #sdcmd_MODE
                    call    #sdcmd_NOP
                    call    #sdcmd_NOP
                    call    #sdcmd_NOP

                    ' preload cache line 0 here
                    ' we can do this here, because only tag line 0 is
                    ' affected, and we are clear of that location

                    mov     addr,#0
                    mov     clineptr,cacheptr   ' cachestart: cache line 0
                    call    #sdramBuffRead      

                    ' We will proceed to clearing the rest of the tag vector, and
                    ' then start the refresh counting loop, and enable
                    ' receiving commands.
                    jmp     #inittagvector

                    ' This is the system clk count for 100us == 0.1ms delay
                    ' But we also rond up to a multiple of 16 clks, so as
                    ' to keep in-step with the phase of the ramclk

initdelayclks       long    (((MY_CLKFREQ + 9_999) / 10_000) + 15) >> 4

'--------------------------------------------------------------------
' SDRAM commands
' use movd later to set commands ... can save andn _DCMD_MASK and toggleClock
'
' Many of these routines are only used in init, so they can be moved into
' the tag vector locations, to save space.
'
toggleClock         mov     phsa,_PHSA          ' edge offset each enable
                    andn    outa,_RAMCLK_MASK      ' turn clock on
                    or      outa,_RAMCLK_MASK      ' turn clock off
toggleClock_ret     ret

sdcmd_MODE          mov     addr,_MODE_REG      ' new board needs address latched on D0-7 too
                    call    #setAddress         ' set address bits for mode
                    andn    outa,_DCMD_MASK     ' mode CMD is 0
                    call    #toggleClock
                    or      outa,_DCMD_NOP      ' send NOP
sdcmd_MODE_ret      ret

sdcmd_NOP           or      outa,_DCMD_NOP      ' send NOP
                    call    #toggleClock
sdcmd_NOP_ret       ret

'--------------------------------------------------------------------
' this set address is used for refresh and startup precharge only.
'
setAddress          or      dira,#DATABUS_MASK    ' dbus output for data or address write
                    mov     tmp, addr
                    and     tmp,_ADDRBUS_MASK
                    andn    outa,_ADDRBUS_MASK
                    or      outa,tmp
                    or      outa,_ALE_MASK
                    andn    outa,_ALE_MASK      ' hold upper address bits
setAddress_ret      ret

'--------------------------------------------------------------------
' Finish the allocation of the tag vector in COG memory:
' This includes all of the init code above, plus more, to get
' to the full vector length.
                    long    0 [MAX_TAGCOUNT-$]

'====================================================================
' mailbox command spinner loop
' user will get a block pointer back and access the block with a mod index such as:
'
'                    mov     index,addr          ' user must do something like this for return pointer array index
'                    and     index,#(LINELEN-1)  ' index can be only up to LINELEN-1


                    ' First, some final initialization code that cannot
                    ' be part of the init code in the tag vector, because
                    ' we are overwriting that vector.
                    '
                    ' In theory, we can reuse some of these locations for
                    ' variables that are used in command dispatch and
                    ' cache address calculations. We will use that
                    ' space optimization when we need it.
                    ' NOT:
                    ' - clineptr
                    ' - addr
                    ' - cacheptr
                    ' - tmp
                    ' - dirty
                    ' - readtag
                    ' - ndx
                    ' YES:
                    ' - saveaddr

inittagvector       mov     tmp, #(MAX_TAGCOUNT - 1)
:inittagloop        movd    :inittagwrite, tmp
                    nop
:inittagwrite       mov     0-0, #0
                    djnz    tmp, #:inittagloop
                    mov     refresh, #1         ' force an immediate refresh
' init is done
                    wrlong  zero,cmdptr         ' let user know we're ready ...
                    ' fall into command loop

' Start command dispatch/refresh loop

sdramDone
cmdloop             djnz    refresh,#norefresh  ' check refresh every time ... djnz here keeps window
                    call    #sdram_refresh      ' if refresh timer expired, reload and refresh

norefresh           rdlong  addr,cmdptr    wz   ' get command/address ... command is in bit 0..1
    if_z            jmp     #cmdloop            ' if zero, do nothing

                    ' Initial refresh check:
                    ' If running through the cache hit/miss check pushes us
                    ' past the time we should perform a refresh, we do one
                    ' early right here.
                    ' This accounts for the time until we get through this
                    ' routine with a cache hit: this is about 23 instructions,
                    ' or about 6 loops.
                    ' sub   refresh, # initial_check_overhead wc, wz
                    sub   refresh, #6   wc, wz
    if_c_or_z       mov   refresh, #0
    if_c_or_z       call  #sdram_refresh

                    mov     clineptr,addr       ' get the cache line pointer offset from address

                    shr     addr,#LINESHIFT wc   ' test for read/write into carry
                    movs    readtag,addr        ' (addr / LINESIZE) mod TAGCOUNT is tag / cache line offset
                    andn    readtag,#$200-TAGCOUNT ' user sends full address. we only load blocks
'                   add     readtag,#8          ' for debugger only
                    muxnc   dirty,_TAG_DIRTY    ' save new tag dirty state - fill pipe slot
readtag             mov     tag,0-0             ' get tag line contents

                    cmpsub  tag,_TAG_DIRTY wc   ' get the old dirty tag
                    cmp     addr,tag       wz
                    muxc    tag,_TAG_DIRTY      ' restore the old dirty tag

    if_e            wrlong  zero,cmdptr         ' if match let user know we're done early ...
                                                ' we rely on the fact that the user is looking for 0 cmdptr
                                                ' and must use another HUB access to get data.
                                                ' problems may come if the user's cog is lower than this cog number

                    and     clineptr,_CLP_MOD   ' user sends full address. we only load blocks
                    add     clineptr,cacheptr   ' get cache line
                    wrlong  clineptr,datptr     ' send cache buffer to user - data may change before we're done

    if_e            jmp     #writeTag
                    call    #flush              ' if bad tag, flush - cache code never changes Z flag
                    wrlong  zero,cmdptr         ' let user know we're ready after flush ...

writeTag            movd    tagit,readtag
                    or      tag,dirty           ' save new tag dirty state
tagit               mov     0-0,tag

                    jmp     #sdramDone          ' 18*4 cycles on cache hit, but 16*4 for delivery.

'--------------------------------------------------------------------
' flush method saves dirty page and loads page from backstore
'
flush               ' if dirty bit set, write old buffer before read
    if_nc           jmp     #flush_rdonly
flush_write
                    ' Check to see if we need to do an early refresh here
                    ' We do a refresh early if the write operation would
                    ' delay the refresh.
                    ' The write refresh loop cost is the number of 16-clk
                    ' iterations consumed by the write code, which should be
                    ' (((CLKS_PER_BYTE*LINELEN) + CLKS_OVHD_WRITE) + 15) >> 4
                    ' sub     refresh, # write_refresh_loop_cost
                    sub     refresh, #((((CLKS_PER_BYTE*LINELEN) + CLKS_OVHD_WRITE) + 15) >> 4) wc, wz
    if_c_or_z       neg     refresh, #((((CLKS_PER_BYTE*LINELEN) + CLKS_OVHD_WRITE) + 15) >> 4)
    if_c_or_z       call    #sdram_refresh

                    mov     saveaddr,addr       ' save addr for read
                    mov     addr,tag            ' get cacheline tag addr
                    shl     addr,#LINESHIFT      ' physical address to write cacheline
                    call    #sdramBuffWrite     ' save cacheline
                    mov     addr,saveaddr       ' restore addr for read
flush_rdonly
                    ' Check to see if we need to do an early refresh here
                    ' We do a refresh early if the read operation would
                    ' delay the refresh.
                    ' The read refresh loop cost is the number of 16-clk
                    ' iterations consumed by the write code, which should be
                    ' (((CLKS_PER_BYTE*LINELEN) + CLKS_OVHD_READ) + 15) >> 4
                    sub     refresh, #((((CLKS_PER_BYTE*LINELEN) + CLKS_OVHD_READ) + 15) >> 4) wc, wz
    if_c_or_z       neg     refresh, #((((CLKS_PER_BYTE*LINELEN) + CLKS_OVHD_READ) + 15) >> 4)
    if_c_or_z       call    #sdram_refresh

                    mov     tag,addr            ' move address clears dirty/count
                    shl     addr,#LINESHIFT      ' physical address to write cacheline
                    call    #sdramBuffRead      ' if zero, load cache line
flush_ret           ret


_TAG_DIRTY          long DIRTYMASK              ' use to mark pages for write-back
_CLP_MOD            long (TAGCOUNT-1)<<LINESHIFT ' use modulo for finding tag address

cmdptr              long 0                      ' cache command pointer
datptr              long 0                      ' cache data pointer
cacheptr            long 0                      ' cache pointer

'====================================================================


'--------------------------------------------------------------------
' refresh
'

' [TD] This is the "new" refresh code:
' It uses a clock rate of 10MHz for the SDRAM clock (FRQA == $2000_0000),
' which gives one clock rising transition per two instruction. So we can
' issue the PRECHARGE command in one instruction, and set the REFRESH
' command two instructions later.
'
' The time spent in this routine is 44 + (8 * REFRESH_BURST_COUNT) clocks.
' But we should include the loop overhead of calling this routine as well,
' which is 8 clks, which gives 52 + (8 * REFRESH_BURST_COUNT) clocks.

sdram_refresh
                    add     refresh, refreshtime ' reset refresh counter
                                                 ' with adjustment from caller

                    ' for safety, we could do this as well:
'                    cmp     refresh, refreshtime wc
'    if_c            mov     refresh, #1

                    or      outa, _SET_A10      ' set A10 for precharge all bank select on new board

                    mov     frqa, _C_2000_0000  ' 8 clocks per cycle: 4 high, 4 low
                    mov     phsa, #0            ' The clock can go low when we do the
                                                ' command write
                                                ' in other words, the low-to-high
                                                ' transition is offset from our write
                                                ' cycle by 4 clocks
                                                ' This is slow enough that we don't
                                                ' need to do NOPs...

                    ' A quick note about timing:
                    ' According to the datasheet, the time from the Precharge
                    ' command to the auto-refresh is at least 18 to 20 ns; here
                    ' we have two full instructions, or 8 clocks, which is fine
                    ' at any possible Propeller clock speed.
                    ' For the time between Refresh commands, the datasheet says
                    ' that at least 60ns is required; here we have two
                    ' instructions, or 8 clocks, which is good up to a
                    ' 133 MHz Propeller clock, which should be good enough.

                    andn    outa,_DCMD_PRECHGNOT' send PRECHARG (change from NOP command)
                    andn    outa,_RAMCLK_MASK      ' four clocks after this write, the precharge
                                                ' command starts
                    mov     irefr, irefr_limit wz
                    xor     outa, _DCMD_PRECHG_XOR_RFRSH
:loop  if_nz        djnz    irefr, #:loop       ' 1, 3, 5, 7, 9... instruction times
                                            ' for 0, 2, 4, 6, 8... irefr_limit counts
                                            ' for 1, 2, 3, 4, 5... refreshes

                    or      outa, _DCMD_NOP
                    or      outa, _RAMCLK_MASK
                    'mov     frqa,_FRQA_FAST    ' already at this speed.

                    andn    outa,_SET_A10       ' release A10 for precharge all bank select
sdram_refresh_ret   ret

' [TD] These values are calculated depending on on the refresh burst length,
' and considering the overhead of the actual refresh routine code itself.

irefr           long  0                   ' inner refresh command loop count.
irefr_limit     long  ((REFRESH_BURST_COUNT - 1) << 1) ' controls how many refreshes to do.

' refresh_time is the number of 16-clock looks we can do before we have to
' perform a refresh.
' See the discussion above for the calculation of the number of loops.

refresh         long  CMDPOLL_LOOPS_PER_REFR_BURST ' initial value
refreshtime     long  CMDPOLL_LOOPS_PER_REFR_BURST


'====================================================================

_LINELEN        long LINELEN                    ' greatest address for cache
_MAX_ABIT       long $0100_0000

_TAG_MASK       long TAGMASK                    ' greatest address for cache
_TAG_MOD        long (TAGCOUNT-1)<<2            ' use modulo for finding tag address
_TAG_OFF        long TAGCOUNT<<2                ' offset for tags
_TAG_LINE_MASK  long (TAGCOUNT-1)<<LINESHIFT     ' use for modulo
_TAG_INC        long $0080_0000
_TAG_CNT        long $FF80_0000
_TAG_CMD        long $C000_0000                 ' used for decoding command
_CLP_OFF        long TAGCOUNT<<LINESHIFT         ' offset of second cache space

'_DATABUS_MASK   long DATABUS_MASK
_CAPIN_MASK     long $0000_03ff
_ADDRBUS_MASK     long ADDRBUS_MASK
_ALE_ADDRBUS_MASK long ADDRBUS_MASK | ALE_MASK
_ALE_MASK       long 1 << ALE_PIN
_RAMCLK_MASK    long RAMCLK_MASK
_WEN_MASK       long 1 << WEN_PIN
_BANK_MASK      long $6000
_ADDR_MASK      long ADDR_MASK
_RASN_MASK      long RASN_MASK
_CASN_MASK      long CASN_MASK
_CPRE_MASK      long CASN_PREPIN_MASK
_DCMD_MASK      long DCMD_MASK                  ' SDRAM COMMAND bits
_DCMD_CLK_MASK  long DCMD_MASK | RAMCLK_MASK

_SET_A10        long $0400  ' A10 used for precharge all banks A7..0 A15..8

_DCMD_NOP       long RASN_MASK | CASN_MASK | WEN_MASK   ' 111 RAS,CAS,WEN
'_DCMD_BTERM     long RASN_MASK | CASN_MASK             ' 110
_DCMD_BTERM_NOT long WEN_MASK                           ' 001
'_DCMD_READ      long RASN_MASK | WEN_MASK              ' 101
_DCMD_READ_NOT  long CASN_MASK                          ' 010
'_DCMD_WRITE     long RASN_MASK                         ' 100
_DCMD_WRITE_NOT long CASN_MASK | WEN_MASK               ' 011
'_DCMD_ACTIVE    long CASN_MASK | WEN_MASK              ' 011
_DCMD_ACTIVENOT long RASN_MASK                          ' 100
'_DCMD_PRECHG    long CASN_MASK                         ' 010
_DCMD_PRECHGNOT long RASN_MASK | WEN_MASK               ' 101
'_DCMD_REFRESH   long WEN_MASK                          ' 001 RAS,CAS,WEN
_DCMD_RFRSH_NOT long RASN_MASK | CASN_MASK              ' 110 RAS,CAS,WEN
_DCMD_MODE      long 0                                  ' 000

_DCMD_PRECHG_XOR_RFRSH long CASN_MASK | WEN_MASK ' (RASN | WEN) ^ (RASN | CASN)

_MODE_REG       long MODE_REG                           ' dont shift on new board

OUTPUT_MASK     long ADDRBUS_MASK| RASN_MASK| CASN_MASK| ALE_MASK| WEN_MASK

' All of these are the same value
_C_8000_0000
_TAG_MSB
_MSB32
_FRQA_FASTER
_PHSA
_PHSA8          long $8000_0000

_C_2000_0000
_FRQA_FAST      long $2000_0000

_C_4000_0000
_PHSA4          long $4000_0000
_PHSA2          long $2000_0000
_PHSA1          long $1000_0000

_CTRA           long (%100_000 << 23) | RAMCLK_PIN

zero            long 0


dirty           long 0
'set2            long 0

'--------------------------------------------------------------------
' the variables that used to be overlayed on the Initialization code
data             long 0
dat1             long 0
dat2             long 0
dat3             long 0
dat4             long 0
dat5             long 0
dat6             long 0
dat7             long 0
bank             long 0
ndx              long 0
tmp              long 0
bufp             long 0
saveaddr         long 0
index            long 0
clineptr         long 0
tags             long 0
tag              long 0
addr             long 0

'--------------------------------------------------------------------

delay2              nop
delay2_ret          ret

'--------------------------------------------------------------------
' Send address sets address into sdram with upper address ACTIVE
' and leaves lower address on the bus for a READ or WRITE to take over.
' This is a burst setup
'
' SDRAM addressing architecture is a little strange. 25 bits for 32MB.
' BA0:1+A0:12 make up the 15 RAS bits. A0:9 make up 10 CAS bits.
' A10 is precharge control, and A11:12 are ignored during CAS cycle.
'
' a full address looks like this B_BHHH_HHhh_hhhh_hhLL_llll_llll
' CBBH_HHHH_hhhh_hhhh shr 10
'

send_ADDRESS
                    or      dira,#DATABUS_MASK    ' set dbus to output 

                    ' setup ROW address
                    mov     tmp,addr            ' %000B_Bhhh_hhhh_hhhh_hhll_llll_llll
                    shr     tmp,#10             ' put hi addr bits on P0-14 %0BBh_hhhh_hhhh_hhhh
                    mov     bank,tmp            ' save bank address bits A13-14 for CAS phase
                    and     bank,_BANK_MASK     ' use only bank bits
                    andn    outa,_ADDRBUS_MASK    ' clear address bits
                    or      outa,tmp            ' send address
                    or      outa,_ALE_MASK      ' send latch address
                    andn    outa,_ALE_MASK      ' address latched

                    ' send RAS ACTIVE - could call #sdcmd_ACTIVE
                    andn    outa,_DCMD_ACTIVENOT' send ACTIVE
                    mov     phsa,_PHSA          ' clock edge offset
                    andn    outa,_RAMCLK_MASK      ' turn on clock
                    nop
                    or      outa,_DCMD_NOP      ' send NOP
                    call    #delay2
                    or      outa,_RAMCLK_MASK      ' turn off clock

                    ' get lower 10 address bits A0..9 ' could be a subroutine
                    mov     tmp,addr            ' %0000_00ll_llll_llll
                    and     tmp,_CAPIN_MASK     ' put low addr bits on P0-14
                    or      tmp,bank            ' include bank bits for read/write
                    andn    outa,_BANK_MASK     ' clear for next write
                    andn    outa,_CAPIN_MASK    ' clear for next write
                    or      outa,tmp            ' set low bits
                    or      outa,_ALE_MASK      ' send latch address
                    andn    outa,_ALE_MASK      ' address latched

                    or      outa,_SET_A10       ' A10 high for full-burst precharge terminate

send_ADDRESS_ret    ret


'--------------------------------------------------------------------
' read Burst ... 32 bytes ...
'
sdramBuffRead       'init and write set dbus to input
                    call    #send_ADDRESS

                    mov     bufp, clineptr      ' load cacheline

                    mov     phsa,_PHSA1         ' edge offset each enable
                    andn    outa,_DCMD_READ_NOT ' set READ from NOP
                    andn    outa,_RAMCLK_MASK      ' turn on clock now for CAS delay time
                    nop
                    or      outa,_RAMCLK_MASK      ' turn on clock now for CAS delay time
                    or      outa,_DCMD_NOP      ' set nops
                    mov     phsa,_PHSA4         ' edge offset each enable
                    andn    outa,_RAMCLK_MASK      ' turn on clock now for CAS delay time
                    andn    dira,#DATABUS_MASK    ' set dbus as input for read

                    mov     phsa,_PHSA4         ' edge offset each enable
                    andn    outa,_RAMCLK_MASK
                    movs    data,ina            ' $0000_0aAA
                    ror     data,#8     wc      ' $AA00_0001 C = LSB
                    movs    data,ina            ' $AA00_0bBB
                    ror     data,#8             ' $BBAA_000b
                    movs    data,ina            ' $BBAA_0cCC
                    ror     data,#17            ' $00cC_BbAa shifted right x 1
                    movi    data,ina            ' $DDcC_BBAa
                    rcl     data,#1             ' $DDCC_BBAA restore LSB
                    movs    dat1,ina            ' $0000_0aAA
                    ror     dat1,#8     wc      ' $AA00_0001 C = LSB
                    movs    dat1,ina            ' $AA00_0bBB
                    ror     dat1,#8             ' $BBAA_000b
                    movs    dat1,ina            ' $BBAA_0cCC
                    ror     dat1,#17            ' $00cC_BbAa shifted right x 1
                    movi    dat1,ina            ' $DDcC_BBAa
                    rcl     dat1,#1             ' $DDCC_BBAA restore LSB
                    movs    dat2,ina            ' $0000_0aAA
                    ror     dat2,#8     wc      ' $AA00_0001 C = LSB
                    movs    dat2,ina            ' $AA00_0bBB
                    ror     dat2,#8             ' $BBAA_000b
                    movs    dat2,ina            ' $BBAA_0cCC
                    ror     dat2,#17            ' $00cC_BbAa shifted right x 1
                    movi    dat2,ina            ' $DDcC_BBAa
                    rcl     dat2,#1             ' $DDCC_BBAA restore LSB
                    movs    dat3,ina            ' $0000_0aAA
                    ror     dat3,#8     wc      ' $AA00_0001 C = LSB
                    movs    dat3,ina            ' $AA00_0bBB
                    ror     dat3,#8             ' $BBAA_000b
                    movs    dat3,ina            ' $BBAA_0cCC
                    ror     dat3,#17            ' $00cC_BbAa shifted right x 1
                    movi    dat3,ina            ' $DDcC_BBAa
                    rcl     dat3,#1             ' $DDCC_BBAA restore LSB
                    movs    dat4,ina            ' $0000_0aAA
                    ror     dat4,#8     wc      ' $AA00_0001 C = LSB
                    movs    dat4,ina            ' $AA00_0bBB
                    ror     dat4,#8             ' $BBAA_000b
                    movs    dat4,ina            ' $BBAA_0cCC
                    ror     dat4,#17            ' $00cC_BbAa shifted right x 1
                    movi    dat4,ina            ' $DDcC_BBAa
                    rcl     dat4,#1             ' $DDCC_BBAA restore LSB
                    movs    dat5,ina            ' $0000_0aAA
                    ror     dat5,#8     wc      ' $AA00_0001 C = LSB
                    movs    dat5,ina            ' $AA00_0bBB
                    ror     dat5,#8             ' $BBAA_000b
                    movs    dat5,ina            ' $BBAA_0cCC
                    ror     dat5,#17            ' $00cC_BbAa shifted right x 1
                    movi    dat5,ina            ' $DDcC_BBAa
                    rcl     dat5,#1             ' $DDCC_BBAA restore LSB
                    movs    dat6,ina            ' $0000_0aAA
                    ror     dat6,#8     wc      ' $AA00_0001 C = LSB
                    movs    dat6,ina            ' $AA00_0bBB
                    ror     dat6,#8             ' $BBAA_000b
                    movs    dat6,ina            ' $BBAA_0cCC
                    ror     dat6,#17            ' $00cC_BbAa shifted right x 1
                    movi    dat6,ina            ' $DDcC_BBAa
                    rcl     dat6,#1             ' $DDCC_BBAA restore LSB
                    movs    dat7,ina            ' $0000_0aAA
                    ror     dat7,#8     wc      ' $AA00_0001 C = LSB
                    movs    dat7,ina            ' $AA00_0bBB
                    ror     dat7,#8             ' $BBAA_000b
                    movs    dat7,ina            ' $BBAA_0cCC
                    ror     dat7,#17            ' $00cC_BbAa shifted right x 1
                    movi    dat7,ina            ' $DDcC_BBAa

                    wrlong  data,bufp
                    add     bufp,#4
                    or      outa,_RAMCLK_MASK
                    wrlong  dat1,bufp
                    add     bufp,#4
                    rcl     dat7,#1             ' $DDCC_BBAA restore LSB
                    wrlong  dat2,bufp
                    add     bufp,#4
                    andn    outa,_DCMD_PRECHGNOT' set PRECHARGE to terminate burst A10 autoprechare ok
                    wrlong  dat3,bufp
                    add     bufp,#4
                    andn    outa,_RAMCLK_MASK      ' turn on clock
                    wrlong  dat4,bufp
                    add     bufp,#4
                    or      outa,_DCMD_NOP      ' send NOP
                    wrlong  dat5,bufp
                    add     bufp,#4
                    or      outa,_RAMCLK_MASK      ' turn off clock
                    wrlong  dat6,bufp
                    add     bufp,#4
                    wrlong  dat7,bufp

sdramBuffRead_ret   ret                         ' 137 instr. for 32 bytes
                                                ' 41I + 32*3I
                                                ' 164 + 32 * 12 clks

'--------------------------------------------------------------------
' write Burst ... 32 bytes
'
sdramBuffWrite
                    mov     bufp, clineptr      ' get all 32 bytes now

                    call    #send_ADDRESS       ' activate address/bank (29 instr. incl. call)

                    rdbyte  data,bufp
                    andn    outa,#$ff
                    or      outa,data           ' First byte for write
                    rdlong  data,bufp
                    add     bufp,#4
                    shr     data,#8             ' get next byte ready
                    rdlong  dat1,bufp
                    add     bufp,#4
                    rdlong  dat2,bufp
                    add     bufp,#4
                    rdlong  dat3,bufp
                    add     bufp,#4
                    rdlong  dat4,bufp
                    add     bufp,#4
                    rdlong  dat5,bufp
                    add     bufp,#4
                    rdlong  dat6,bufp
                    add     bufp,#4
                    andn    outa,_DCMD_WRITE_NOT' set WRITE command from NOP
                    rdlong  dat7,bufp

                    mov     phsa,_PHSA8         ' edge offset each enable
                    andn    outa,_RAMCLK_MASK      ' turn on clock now for CAS delay time
                    mov     phsa,_PHSA4         ' clock edge offset
                    or      outa,_DCMD_NOP      ' set NOP after write

                    movs    outa,data           ' A0 has no impact after WRITE cmd
                    shr     data,#8
                    movs    outa,data           
                    shr     data,#8
                    movs    outa,data           
                    nop
                    movs    outa,dat1           ' next byte for write
                    shr     dat1,#8
                    movs    outa,dat1           
                    shr     dat1,#8
                    movs    outa,dat1           
                    shr     dat1,#8
                    movs    outa,dat1           
                    nop
                    movs    outa,dat2           ' next byte for write
                    shr     dat2,#8
                    movs    outa,dat2           
                    shr     dat2,#8
                    movs    outa,dat2           
                    shr     dat2,#8
                    movs    outa,dat2           
                    nop
                    movs    outa,dat3           ' next byte for write
                    shr     dat3,#8
                    movs    outa,dat3           
                    shr     dat3,#8
                    movs    outa,dat3           
                    shr     dat3,#8
                    movs    outa,dat3           
                    nop
                    movs    outa,dat4           ' next byte for write
                    shr     dat4,#8
                    movs    outa,dat4           
                    shr     dat4,#8
                    movs    outa,dat4           
                    shr     dat4,#8
                    movs    outa,dat4           
                    nop
                    movs    outa,dat5           ' next byte for write
                    shr     dat5,#8
                    movs    outa,dat5           
                    shr     dat5,#8
                    movs    outa,dat5           
                    shr     dat5,#8
                    movs    outa,dat5           
                    nop
                    movs    outa,dat6           ' next byte for write
                    shr     dat6,#8
                    movs    outa,dat6           
                    shr     dat6,#8
                    movs    outa,dat6           
                    shr     dat6,#8
                    movs    outa,dat6           
                    nop
                    movs    outa,dat7           ' next byte for write
                    shr     dat7,#8
                    movs    outa,dat7           
                    shr     dat7,#8
                    movs    outa,dat7           
                    shr     dat7,#8
                    movs    outa,dat7           
                    or      outa,_RAMCLK_MASK   ' 132 instr. to here

                    call    #burstTerminate
sdramBuffWrite_ret  ret                         ' 143 instructions for 32 bytes (3 per byte)
                                                ' 47I + 32*3I
                                                ' 188 + 32 * 12 clks

burstTerminate
                    andn    outa,_DCMD_PRECHGNOT' set PRECHARGE to terminate burst A10 autoprechare ok
                    andn    outa,_RAMCLK_MASK      ' turn on clock
                    mov     phsa,_PHSA4         ' clock offset guarantees clock for BURST TERM
                    or      outa,_DCMD_NOP      ' send NOP
                    call    #delay2
                    or      outa,_RAMCLK_MASK      ' turn off clock

burstTerminate_ret  ret

'====================================================================

                    ' For the current code, there are 11 longs free here in the cog image
fillpasm            long 0 [$1ef-$]
endpasm             fit $1ef
